/*
 * Copyright (C) 2013 Philippe Komma
 * 
 * This file is part of Cartoon: A video stream cartoonizer 
 * 
 * Cartoon is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Foobar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Foobar. If not, see <http://www.gnu.org/licenses/>.
 */

#include "Cartoon.h"

#include <iostream>
	using std::endl;
	using std::cout;

Cartoon::Cartoon() : 
	m_subW( 40 ),
	m_subH( 30 ),
	m_numc( 12 ),
	m_imgGray( NULL ),
	m_imgSobelHor( NULL ),
	m_imgSobelVer( NULL ),
	m_imgLaplace( NULL ),
	m_imgResult( NULL ),
	m_imgSmoothed( NULL ),
	m_matResampled( NULL ),
	m_matLabels( NULL ),
	m_matCenters( NULL )
{
	CvSize imgSize = { 640, 480 };

	m_imgGray = cvCreateImage( imgSize, IPL_DEPTH_8U, 1 );

	m_imgSobelHor = cvCreateImage( imgSize, IPL_DEPTH_32F, 1 );
	m_imgSobelVer = cvCreateImage( imgSize, IPL_DEPTH_32F, 1 );

	m_imgLaplace = cvCreateImage( imgSize, IPL_DEPTH_32F, 1 );

	m_imgResult = cvCreateImage( imgSize, IPL_DEPTH_8U, 3 );

	m_imgSmoothed = cvCreateImage( imgSize, IPL_DEPTH_8U, 3 );

	m_matResampled = cvCreateMat( m_subW * m_subH, 3, CV_32F );

	m_matLabels = cvCreateMat( m_subW * m_subH, 1, CV_32S );

	m_matCenters = cvCreateMat( m_numc, 3, CV_32F );
}

IplImage* Cartoon::exec( IplImage* img )
{
	cvCvtColor( img, m_imgGray, CV_RGB2GRAY );

	cvSobel( m_imgGray, m_imgSobelHor, 1, 0, 9 );
	cvSobel( m_imgGray, m_imgSobelVer, 0, 1, 9 );

	float* srcSobelHor = ( float* )m_imgSobelHor->imageData;
	float* srcSobelVer = ( float* )m_imgSobelVer->imageData;

	float* ptrLaplace = ( float* )m_imgLaplace->imageData;

	int x, y;
	int offset = 0;
	float maxv = -1;		
	for ( y = 0; y < m_imgLaplace->height; y++ ) {
		for ( x = 0; x < m_imgLaplace->width; x++ ) {
			float aux = sqrtf( srcSobelHor[ offset ] * srcSobelHor[ offset ] +
				srcSobelVer[ offset ] * srcSobelVer[ offset ] );

			ptrLaplace[ offset ] = aux;

			offset++;

			maxv = aux > maxv ? aux : maxv;				
		}
	}

	unsigned char* ptrSource = ( unsigned char* )img->imageData;		

	float* ptrResampled = m_matResampled->data.fl;

	for ( y = 0; y < m_imgResult->height; y++ ) {
		for ( x = 0; x < m_imgResult->width; x++ ) {
			if ( ( x & 15 ) == 0 && ( y & 15 ) == 0 ) {
				*( ptrResampled     ) = ( float )*( ptrSource     );
				*( ptrResampled + 1 ) = ( float )*( ptrSource + 1 );
				*( ptrResampled + 2 ) = ( float )*( ptrSource + 2 );				

				ptrResampled += 3;				
			}

			ptrSource += 3;
		}
	}

	CvTermCriteria crit;

	crit.epsilon = 1e-2;
	crit.max_iter = 5;
	crit.type = CV_TERMCRIT_EPS | CV_TERMCRIT_ITER;		

	cvKMeans2( m_matResampled, m_numc, m_matLabels, crit, 1, 0, cv::KMEANS_RANDOM_CENTERS, m_matCenters );

	ptrSource = ( unsigned char* )img->imageData;		

	unsigned char* ptrResult = ( unsigned char* )m_imgResult->imageData;

	for ( y = 0; y < m_imgResult->height; y++ ) {
		for ( x = 0; x < m_imgResult->width; x++ ) {
			int j;
			int mini = -1;
			float minv = FLT_MAX;
			for ( j = 0; j < m_numc; j++ ) {
				int c;
				float sum = 0;
				for ( c = 0; c < 3; c++ ) {
					const float u = *( m_matCenters->data.fl + j * 3 + c );
					const float v = *( ptrSource + c );
					sum += ( u - v ) * ( u - v );
				}
				if ( sum < minv ) {
					mini = j;
					minv = sum;
				}
			}

			*( ptrResult     ) = ( unsigned char )( *( m_matCenters->data.fl + mini * 3     ) );
			*( ptrResult + 1 ) = ( unsigned char )( *( m_matCenters->data.fl + mini * 3 + 1 ) );
			*( ptrResult + 2 ) = ( unsigned char )( *( m_matCenters->data.fl + mini * 3 + 2 ) );

			ptrSource += 3;

			ptrResult += 3;
		}
	}

	cvSmooth( m_imgResult, m_imgSmoothed );

	ptrResampled = m_matResampled->data.fl;

	unsigned char* ptrSmoothed = ( unsigned char* )m_imgSmoothed->imageData;

	for ( y = 0; y < m_imgSmoothed->height; y++ ) {
		for ( x = 0; x < m_imgSmoothed->width; x++ ) {
			float aux = *ptrLaplace / maxv;

			ptrLaplace++;

			aux = aux > 0.075f ? 0.25f : 1;

			if ( ( y < m_subH ) && ( x < m_subW ) ) {
				*( ptrSmoothed     ) = ( unsigned char )( *( ptrResampled     ) );
				*( ptrSmoothed + 1 ) = ( unsigned char )( *( ptrResampled + 1 ) );
				*( ptrSmoothed + 2 ) = ( unsigned char )( *( ptrResampled + 2 ) );					
				ptrResampled += 3;
			} else {
				*( ptrSmoothed     ) = ( unsigned char )( *( ptrSmoothed     ) * aux );
				*( ptrSmoothed + 1 ) = ( unsigned char )( *( ptrSmoothed + 1 ) * aux );
				*( ptrSmoothed + 2 ) = ( unsigned char )( *( ptrSmoothed + 2 ) * aux );
			}

			ptrSmoothed += 3;
		}
	}

	return m_imgSmoothed;

	return NULL;
}